import numpy as np


class PopulationPlotter:
    def __init__(self, pop):
        self._pop = pop

    def variable_pair(self, x, y, **kw):
        """
        Scatter plot of two variables.

        Parameters
        ----------
        :x:     `str` for the X axis column.
        :y:     `str` for the Y axis column.

        Key-Word Arguments - Pandas DataFrame Scatter Plot K-W Arguments
        """
        df = self._pop.datasets.output
        return df.plot.scatter(x=x, y=y, **kw)

    def history(self, column, metric='mean', secondary_metric=None, **kw):
        """
        Scatter plot of the evolution of through multiple generations.

        Parameters
        ----------
        :column:            `str` for the Y axis column.
        :metric:            `str` or `function`
        :secondary_metric:  `str` or `function`

        Key-Word Arguments - Pandas DataFrame Scatter Plot K-W Arguments
        """
        metric = set_metric(metric)
        gcol = 'generation'
        df = self._pop.datasets.history.groupby(gcol).agg(metric).reset_index()
        ax = df.plot.scatter(x=gcol, y=column, **kw)
        if secondary_metric is not None:
            secondary_metric = set_metric(secondary_metric)
            df = self._pop.datasets.history.groupby(gcol).agg(secondary_metric)\
                .reset_index()
            ax = df.plot.scatter(x=gcol, y=column, ax=ax, color='r', **kw)
        return ax


def set_metric(metric):
    if isinstance(metric, str):
        if metric.lower() == 'min':
            metric = min
        elif metric.lower() == 'max':
            metric = max
        elif metric.lower() in ['mean', 'average', 'avg']:
            metric = np.mean
    return metric


class PopulationPlotterPareto(PopulationPlotter):
    def fronts(self, x=None, y=None, hlfront=None, hlcolor='k',
               show_worst=False, colormap='rainbow', **kw):
        """
        Plot fronts in a 2D scatter plot the separates each front by color and
        highlight one specific front (if desired).

        Parameters
        ----------
        :x:             `str` for the X axis column (Default: None; if None is set,
                        the first fitness column will be used).
        :y:             `str` for the Y axis column (Default: None; if None is set,
                        the second fitness column will be used).
        :hlfront:       `int` or `list` for setting the highlighted front(s)
                        (default: None; if None is set, no fronts will be highlighted)
        :hlcolor:       `str` or `list` for setting the color of the highlighted front
                        (default: 'k')
        :show_worst:    `bool` set True to show the elements that didn't make the cut
                        (default: False)
        :colormap:      `str` for the desired colormap (default: 'rainbow')

        Key-Word Arguments - Pandas DataFrame Scatter Plot K-W Arguments
        """
        df = self._pop.datasets.fitness
        x, y = set_xy(x, y, self._pop.datasets._fitnesscols)
        dfplot = make_dfplot(df, show_worst)
        ax = dfplot.plot.scatter(x=x, y=y,
                                 c='front',
                                 colormap=colormap,
                                 sharex=False,
                                 **kw)
        hlfront = set_hlfront(hlfront)
        if hlfront is not None:
            hlcolor = set_hlcolor(hlfront, hlcolor)
            ax = highlight_fronts(dfplot, x, y, hlfront, hlcolor, ax)
        return ax


def set_xy(x, y, fitcols):
    if x is None:
        x = fitcols[0]
    if y is None:
        y = fitcols[1]
    return x, y


def make_dfplot(df, show_worst):
    if show_worst == False:
        worst = df['front'].max()
        f = df['front'] != worst
        return df[f]
    else:
        return df


def set_hlfront(hlfront):
    if isinstance(hlfront, int):
        hlfront = [hlfront]
    return hlfront


def set_hlcolor(hlfront, hlcolor):
    if isinstance(hlcolor, str):
        return [hlcolor] * len(hlfront)
    elif len(hlcolor) == len(hlfront):
        return hlcolor
    else:
        s = f"""
        Number of `hlfront`s and `hlcolor`s differ.
        (hlfronts: {hlfront} | hlcolors: {hlcolor})
        """
        raise Exception(s)


def highlight_fronts(df, x, y, hlfronts, hlcolors, ax):
    fronts = df['front']
    for front, color in zip(hlfronts, hlcolors):
        ax = df[fronts == front]\
            .plot\
            .scatter(x=x, y=y, ax=ax, color=color)
    return ax



from copy import deepcopy
import numpy as np
import pandas as pd
from lamarck import Creature
from lamarck.fitness import FitnessBuilder
from lamarck.repopulate import Repopulator
from lamarck.plotter import PopulationPlotter
from lamarck.utils import (create_id, genome_already_exists,
                           is_objective_ascending)


class Population:
    """
    Populations are collections of genes in Pandas DataFrames. The basic flow of
    Population objects in GA Optimizations is:

        1- Define the Genome
        2- Create a Population based on the Genome
        3- Simulate the Population on an Environment
        4- Select the Fittest and eliminate the weaker Creatures
        5- Generate Offspring
        6- Mutate
        7- IF (Converged) THEN proceed ELSE Go To (3)
        8- End

    Example
    -------
    >>> pop_creator = PopulationCreator(genome_blueprint)
    # (step 2)
    >>> pop = pop_creator.create.det(n)
    # will create a population of (number of genes)**n Creatures
    # (step 3)
    >>> env = Environment()
    >>> env.set_process(the_function)
    >>> env.simulate(pop)
    # "pop" will now store the results of the simulations
    >>> pop\
        .apply_fitness\
        .multi_objective\
        .pareto(outputs=['a', 'b'], objective=['min', 'max'])
    # (step 4)
    >>> selected_pop = pop.select(p=0.5)
    # (step 5)
    >>> selected_pop.reproduce.tournamet()
    # (step 5)
    >>> selected_pop.reproduce.mutate()
    """

    def __init__(self, genome_blueprint):
        self.genome_blueprint = genome_blueprint
        self.generation = 0
        self.genes = list(genome_blueprint)
        self.natural_population_level = None
        self.fitness_rank_method = None
        # instances
        self.populate = Populator(self)
        self.datasets = PopulationDatasets(self)
        self.reproduce = Repopulator(self)
        self.get_creature = CreatureGetter(self)
        self.apply_fitness = FitnessBuilder(self)
        self._set_plotter(PopulationPlotter)

    def __getitem__(self, key):
        return self.get_creature.from_index(key)

    def __len__(self):
        return len(self.datasets.input)

    def __repr__(self):
        specs = [f"{x} ({d['type']})"
                 for x, d in self.genome_blueprint.items()]
        genestr = ', '.join(specs)
        return f"""Population with {len(self)} Creatures with genes {genestr}.
        """

    def __add__(self, other):
        genome_blueprint = deepcopy(self.genome_blueprint)
        df = pd\
            .concat((self.datasets.input, other.datasets.input))\
            .reset_index(drop=True)
        new_pop = Population(genome_blueprint)
        new_pop.populate.from_genome_dataframe(df)
        new_pop.datasets._drop_duplicates()
        return new_pop

    def _set_plotter(self, plotter_class):
        self.plot = plotter_class(self)
        self._plotter_class = plotter_class

    def copy(self):
        genome_blueprint = deepcopy(self.genome_blueprint)
        new_pop = Population(genome_blueprint)
        new_pop.datasets._absorb(self.datasets)
        new_pop.natural_population_level = self.natural_population_level
        new_pop.generation = self.generation
        new_pop.fitness_rank_method = self.fitness_rank_method
        new_pop._set_plotter(self._plotter_class)
        new_pop.apply_fitness._constraints = deepcopy(
            self.apply_fitness._constraints)
        return new_pop

    def define(self, n=None):
        """
        Define the Population's "normal" size.

        Parameters
        ----------
        :n:     `int` representing the size. If `None`, the current population
                size will be considered the norm (default: `None`).
        """
        if n is None:
            self.natural_population_level = len(self)
        else:
            self.natural_population_level = n

    def save_to_history(self):
        """
        Appends the fitness dataset to the history dataset.
        """
        if self.datasets.fitness is not None:
            if self.generation not in self.datasets._generations:
                self.datasets._add_to_history(self.datasets.fitness)
            else:
                raise Exception("This generation already exists in history.")
        else:
            raise Exception("Fitness dataset not present.")

    def remove_from_history(self, generation):
        """
        Removes a generation from the history dataset.
        """
        f = self.datasets.history.generation != generation
        new_history = self.datasets.history[f]
        self.datasets._update_history(new_history)

    def set_outputs(self, outputs):
        """
        Parameters
        ----------
        :outputs:       `list`
        """
        self.datasets._set_outputs(outputs)

    def add_output(self, output_df):
        """
        Parameters
        ----------
        :output_df:     `DataFrame`
        """
        self.datasets._add_output(output_df)

    def set_fitness(self, fitness_df, objectives, rank_method,
                    fitness_cols=None):
        """
        Parameters
        ----------
        :fitness_df:    `DataFrame`
        :objectives:    `list`
        :rank_method:   `function`
        :fitness_cols:  `list` (Default: None)
        """
        self.datasets._set_fitness(fitness_df, objectives, fitness_cols)
        self.fitness_rank_method = rank_method

    @property
    def fitness_rank(self):
        if self.fitness_rank_method is not None:
            return self.fitness_rank_method(self.datasets.fitness)

    def reset_fitness_objectives(self, fitness_cols, objectives):
        """
        Parameters
        ----------
        :fitness_cols:  `list`
        :objectives:    `list`
        """
        self.datasets._set_fitness_objectives(fitness_cols, objectives)

    def select(self, p=0.5):
        """
        Parameters
        ----------
        :p:     `float`
        """
        self.datasets._select(p)
        self.set_generation(self.generation + 1)

    def set_generation(self, generation):
        """
        Parameters
        ----------
        :generation:     `int`
        """
        self.generation = generation


class Populator:
    def __init__(self, population):
        self._pop = population

    def creature(self, creature):
        """
        Adds a Creature to the Population.

        Parameters
        ----------
        :creature:  `Creature` that will be added to the Population.
        """
        self.from_genome(creature.genome)

    def from_genome(self, genome):
        """
        Adds a Creature to the Population, based on its Gene.

        Parameters
        ----------
        :genome:  `dict` representing the `Creature`'s genome.
        """
        raw_genome_df = pd.DataFrame({gene: [v] for gene, v in genome.items()})
        genome_df = create_id(raw_genome_df)
        self._pop.datasets._add_creatures(genome_df)

    def from_creature_list(self, creature_list):
        """
        Adds all Creatures in a `list`.

        Parameters
        ----------
        :creature_list: `list` of Creatures that will be added to the Population.
        """
        for creature in creature_list:
            self.creature(creature)

    def from_genome_list(self, genome_list):
        """
        Adds all Creatures Genomes in a `list`.

        Parameters
        ----------
        :genome_list:   `list` of Genomes that will become the Creatures that will
                        be added to the Population.
        """
        genes = genome_list[0].keys()
        genome_data = {gene: [g[gene] for g in genome_list] for gene in genes}
        raw_genome_df = pd.DataFrame(genome_data)
        genome_df = create_id(raw_genome_df)
        self._pop.datasets._add_creatures(genome_df)

    def from_genome_dataframe(self, dataframe):
        """
        Populates this Population object from a Pandas DataFrame with columns that
        coincide with the Genome Blueprint's Genes and Types.

        Example
        -------
        >>> genome_blueprint
        {'x': {'type': 'numeric',
         'domain': 'int',
         'ranges': {'min': 0, 'max': 10, 'progression': 'linear'}},
        'y': {'type': 'categorical', 'domain': ['A', 'B', 'C']},
        'z': {'type': 'vectorial',
         'domain': ['i', 'j', 'k'],
         'ranges': {'length': 3, 'replace': False}}}

        >>> df
           x  y          z
        0  0  A  (i, j, k)
        1  1  A  (j, i, k)
        2  0  B  (k, i, j)
        3  2  B  (i, j, k)
        4  7  C  (j, i, k)

        >>> pop = Population(genome_blueprint)
        >>> pop.populate.from_genome_dataframe(df)
        """
        self._pop.datasets._set(dataframe.copy())


class PopulationDatasets:
    def __init__(self, population):
        self._pop = population
        self._outputcols = None
        self._fitnesscols = None
        self._objectives = []
        self.input = None
        self.output = None
        self.fitness = None
        self.history = None

    @property
    def index(self): return self.input.index

    @property
    def _inputcols(self): return self._pop.genes

    @property
    def _generations(self):
        if self.history is None:
            return []
        else:
            return self.history.generation.unique().tolist()

    def _set(self, df):
        self.input = create_id(df[self._inputcols])
        self._drop_duplicates()
        self.output = self.input.copy()
        self.fitness = self.input.copy()

    def _set_outputs(self, outputs):
        if self._outputcols is None:
            self._outputcols = outputs
            size = len(self._pop)
            outcols = {}
            for output in outputs:
                col = np.empty(size, dtype=float)
                col.fill(np.nan)
                outcols.update({output: col})
            self.output = self.output.assign(**outcols)

    def _add_output(self, output_df):
        self.output.update(output_df, overwrite=False)

    def _add_creatures(self, genome_df):
        self.input = pd.concat((self.input, genome_df))
        self._update_datasets()

    def _update_datasets(self):
        self._drop_duplicates()
        self.output = pd\
            .merge(self.input, self.output, 'left')\
            .set_index(self.index)
        self.fitness = pd\
            .merge(self.input, self.fitness, 'left')\
            .set_index(self.index)

    def _drop_duplicates(self):
        self.input = self.input.drop_duplicates()

    def _set_fitness(self, fitness_df, objectives, fitness_cols=None):
        if fitness_cols is None:
            fitness_cols = list(fitness_df.columns)
        self._set_fitness_objectives(fitness_cols, objectives)
        self.fitness = pd.concat(
            (self.output, fitness_df), join='inner', axis=1)
        self._sort_fitness()

    @property
    def _ascending(self):
        return [is_objective_ascending(objective)
                for objective in self._objectives]

    def _sort_fitness(self):
        if self._fitnesscols is not None:
            self.fitness.sort_values(self._fitnesscols,
                                     ascending=self._ascending,
                                     inplace=True)

    def _set_fitness_objectives(self, cols, objectives):
        self._fitnesscols = cols
        self._objectives = objectives

    def _add_to_history(self, fitness_df):
        fitness_df_gen = fitness_df.assign(generation=self._pop.generation)
        if self.history is None:
            self.history = fitness_df_gen
        else:
            self.history = pd.concat((self.history, fitness_df_gen))

    def _update_history(self, history_df):
        self.history = history_df

    def _sort_history(self):
        sortcols = self._fitnesscols + ['generation']
        ascending = self._ascending + [True]
        self.history.sort_values(sortcols, ascending=ascending, inplace=True)

    def _absorb(self, other):
        self._outputcols = other._outputcols
        self._fitnesscols = other._fitnesscols
        self._objectives = other._objectives
        self.input = other.input
        self.output = other.output
        self.fitness = other.fitness
        self.history = other.history

    def _select(self, p):
        if self.fitness is None:
            raise Exception(
                "Population wasn't tested yet (Fitness Dataset does't exist)."
            )
        else:
            if self._pop.natural_population_level is None:
                npop = len(self.fitness)
            else:
                npop = self._pop.natural_population_level
            self._add_to_history(self.fitness)
            n_fittest = int(npop * p)
            index = self.fitness[0:n_fittest].index
            self.input = self.input.loc[index]
            self.output = self.output.loc[index]
            self.fitness = self.fitness.loc[index]

    def get_generation_from_history(self, generation):
        f = self.history.generation == generation
        return self.history[f]


class CreatureGetter:
    def __init__(self, population):
        self._pop = population

    def from_id(self, idval):
        """
        Get Creature from the Input Dataset based on its unique ID.
        """
        df = self._pop.datasets.input
        row = df.loc[idval]
        return make_creature_from_df_row(row, self._pop.genes)

    def from_index(self, index):
        """
        Get Creature from the Input Dataset based on its position index.
        """
        df = self._pop.datasets.input
        row = df.iloc[index]
        return make_creature_from_df_row(row, self._pop.genes)

    def best(self):
        """
        Get the Best Creature(s) from the Fitness Dataset.
        """
        f = self._pop.fitness_rank == 1
        ids = self._pop.fitness_rank[f].index
        df_best = self._pop.datasets.input.loc[ids]
        return [make_creature_from_df_row(row, self._pop.genes)
                for _, row in df_best.iterrows()]


def make_creature_from_df_row(row, genes):
    genome = {gene: val for gene, val in row.items() if gene in genes}
    return Creature(genome)



import itertools
import numpy as np
import pandas as pd


def idfunc(x):
    return hash(x)


def column_aggregator(df):
    for col in df.columns:
        yield df[col]


def create_id(df):
    df = df.copy()
    colgen = column_aggregator(df)
    cols = tuple(colgen)
    idcol = pd.Series(tuple(zip(*cols))).apply(idfunc)
    return df.assign(id=idcol).set_index('id')


def get_id(genome):
    genomevals = tuple(genome.values())
    return idfunc(genomevals)


def genome_already_exists(genome, pop):
    return get_id(genome) in pop.datasets.index


def creature_already_exists(creature, pop):
    return creature.id in pop.datasets.index


def is_objective_ascending(objective):
    if objective.lower() in ['min', 'minimize']:
        ascending = True
    elif objective.lower() in ['max', 'maximize']:
        ascending = False
    else:
        raise Exception(":objective: must be either 'min' or 'max'.")
    return ascending


def compare_to_value(val1, oper, val2):
    if oper.lower() == 'eq':
        return val1 == val2
    elif oper.lower() == 'neq':
        return val1 != val2
    elif oper.lower() == 'lt':
        return val1 < val2
    elif oper.lower() == 'le':
        return val1 <= val2
    elif oper.lower() == 'gt':
        return val1 > val2
    elif oper.lower() == 'ge':
        return val1 >= val2
    else:
        raise Exception(f'Invalid operator "{oper}".')


def deterministic_linear(start, stop, n, dtype):
    return np.linspace(start, stop, n, dtype=dtype)


def deterministic_log(start, stop, n, dtype):
    return np.geomspace(start, stop, n, dtype=dtype)


def vectorial_distribution(n, length, domain):
    perm_gen = itertools.product(domain, repeat=length)
    end = len(domain) ** length
    step = int(np.ceil(end / n))
    dist_gen = itertools.islice(perm_gen, 0, end, step)
    return tuple(dist_gen)


def vectorial_distribution_set(n, length, domain):
    end = np.math.perm(len(domain), length)
    if end > 1e9:  # if too many permutations, randomize to avoid performance issues
        dist = tuple(make_vector_list(n, length, domain, False))
    else:
        perm_gen = itertools.permutations(domain, length)
        step = int(np.ceil(end / n))
        dist_gen = itertools.islice(perm_gen, 0, end, step)
        dist = tuple(dist_gen)
    return dist


def random_linear(start, stop):
    return np.random.uniform(start, stop)


def random_linear_dist(start, stop, n, dtype):
    return np.random.uniform(start, stop, n).astype(dtype)


def random_log_dist(start, stop, n, dtype):
    dist = [random_log(start, stop) for _ in n]
    return np.array(dist, dtype=dtype)


def random_log(start, stop):
    lstart = np.log(start)
    lstop = np.log(stop)
    val = np.random.uniform(lstart, lstop)
    return np.exp(val)


def make_vector_list(n, length, domain, replace):
    return [tuple(np.random.choice(domain, length, replace=replace))
            for _ in range(n)]



import itertools
import numpy as np
import pandas as pd
from lamarck import Population
from lamarck.utils import (random_linear_dist,
                           random_log_dist,
                           deterministic_linear,
                           deterministic_log,
                           vectorial_distribution,
                           vectorial_distribution_set,
                           make_vector_list)


class PopulationCreator:
    def __init__(self, genome_blueprint=None):
        self.create = Creator(self)
        self.set_genome_blueprint(genome_blueprint)

    def set_genome_blueprint(self, genome_blueprint):
        self.genome_blueprint = genome_blueprint


class Creator:
    def __init__(self, pop_creator):
        self._pop_creator = pop_creator

    @property
    def _bp(self):
        return self._pop_creator.genome_blueprint

    def det(self, n):
        """
        Create Gene Values Deterministically, distributed according to its Genome
        Blueprint.

        Parameters
        ----------
        :n:     `int` that set the amount of values for every gene. 

        Examples
        --------
        >>> genome_blueprint = {
            'x': {
                'type': 'numeric',
                'domain': 'int',
                'ranges': {
                    'min': 0,
                    'max': 10,
                    'progression': 'linear'}},
            'y': {
                'type': 'categorical',
                'domain': ['i', 'j', 'k']},
            'z': {
                'type': 'vectorial',
                'domain': ['A', 'B', 'C', D, E],
                'ranges': {
                    'length': 3,
                    'replace': False}}
        }
        >>> creator = PopulationCreator(genome_blueprint)
        >>> creator.create.det(5)
        Will create a population with:
        - x values equal to:    [0, 2, 5, 7, 10]
        - y values equal to:    ['i', 'j', 'k']
        - z values equal to:    [('A', 'B', 'C'), ('B', 'A', 'E'), ('C', 'B', 'E'),
                                 ('D', 'C', 'E'), ('E', 'D', 'C')]
        which totalizes 5*3*5 = 75 different genes, or 75 different Creatures in the
        Population.
        """
        if isinstance(n, int):
            return deterministically_make_population(n, self._bp)
        else:
            raise TypeError('Parameter `n` must be `int`.')

    def det_custom(self, ndict):
        """
        Create Gene Values Deterministically, distributed according to its Genome
        Blueprint.

        Parameters
        ----------
        :ndict: `dict` the specify the amount per gene.

        Examples
        --------
        >>> genome_blueprint = {
            'x': {
                'type': 'numeric',
                'domain': 'int',
                'ranges': {
                    'min': 0,
                    'max': 10,
                    'progression': 'linear'}},
            'y': {
                'type': 'categorical',
                'domain': ['i', 'j', 'k']},
            'z': {
                'type': 'vectorial',
                'domain': ['A', 'B', 'C', D, E],
                'ranges': {
                    'length': 3,
                    'replace': False}}
        }
        >>> creator = PopulationCreator(genome_blueprint)
        >>> creator.create.det_custom({'x': 6, 'y': 2})
        Will create a population with:
        - x values equal to:    [0, 2, 4, 6, 8, 10]
        - y values equal to:    ['i', 'k']
        - no values for z,
        which totalizes 6*2 = 12 different genes, or 10 different Creatures in the
        Population. 
        """
        if isinstance(ndict, int):
            return deterministically_make_population_from_dict(ndict, self._bp)
        else:
            raise TypeError('Parameter `ndict` must be `int`.')

    def rand(self, n):
        """
        Create Gene Values Randomly, distributed according to its Genome Blueprint.
        Parameters
        ----------
        :n:     `int` that set the amount of genes that will be made into Creatures in
                the Population.

        Examples
        --------
        >>> >>> genome_blueprint = {
            'x': {
                'type': 'numeric',
                'domain': 'int',
                'ranges': {
                    'min': 0,
                    'max': 10,
                    'progression': 'linear'}},
            'y': {
                'type': 'categorical',
                'domain': ['i', 'j', 'k']},
            'z': {
                'type': 'vectorial',
                'domain': ['A', 'B', 'C', D, E],
                'ranges': {
                    'length': 3,
                    'replace': False}}
        }
        >>> creator = PopulationCreator(genome_blueprint)
        >>> creator.create.rand(50)
        Will create a population with 50 members, with the following values for the
        genes:
        - x values equal to:    any integer between 0 and 10
        - y values equal to:    any object from the list ['i', 'j', 'k']
        - z values equal to:    any sequence of 3 itens from all possible permutations
                                of ['A', 'B', 'C', D, E]
        """
        if isinstance(n, int):
            return randomly_make_population(n, self._bp)
        else:
            raise TypeError('Parameter `n` must be `int`.')


# for deterministic numeric distribution

def deterministically_make_population(n, blueprint):
    if n > 0:
        n_dict = {name: n for name in blueprint}
        return deterministically_make_population_from_dict(n_dict, blueprint)
    else:
        return Population(blueprint)


def deterministically_make_population_from_dict(ndict, blueprint):
    gene_names = []
    gene_vals = []
    for gene_name, specs in blueprint.items():
        n = ndict[gene_name]
        gene_names.append(gene_name)
        vals = get_vals_from_blueprint(n, specs)
        gene_vals.append(vals)
    genes = itertools.product(*gene_vals)
    gene_df = pd.DataFrame(genes, columns=gene_names)
    pop = Population(blueprint)
    pop.populate.from_genome_dataframe(gene_df)
    return pop


def get_vals_from_blueprint(n, specs):
    gene_type = specs['type']
    domain = specs['domain']
    if gene_type == 'numeric':
        ranges = specs['ranges']
        return generate_numeric_distribution(n, ranges, domain)
    elif gene_type == 'categorical':
        return generate_categorical_distribution(n, domain)
    elif gene_type == 'vectorial':
        ranges = specs['ranges']
        return generate_vectorial_distribution(n, ranges, domain)
    else:
        raise Exception(
            'Gene type must be "numeric", "categorical" or "vectorial"'
            f' but instead got {gene_type}.')


def generate_numeric_distribution(n, ranges, domain):
    start = ranges['min']
    stop = ranges['max']
    progression = ranges['progression']
    dtype = get_dtype(domain)
    if progression == 'linear':
        return deterministic_linear(start, stop, n, dtype)
    elif progression == 'log':
        return deterministic_log(start, stop, n, dtype)
    else:
        raise Exception(
            'Gene progression must be "linear" or "log"'
            f' but instead got {progression}.')


def generate_categorical_distribution(n, domain):
    end = len(domain) - 1
    indexes = np.linspace(0, end, n, dtype=int)
    return np.array(domain)[indexes]


def generate_vectorial_distribution(n, ranges, domain):
    length = ranges['length']
    replace = ranges['replace']
    if replace:
        return vectorial_distribution(n, length, domain)
    else:
        return vectorial_distribution_set(n, length, domain)


def get_dtype(domain):
    if domain == 'int':
        return int
    elif domain == 'float':
        return float


# for randomly numeric distribution


def randomly_make_population(n, blueprint):
    if n > 0:
        genes = {gene_name: get_random_vals_from_blueprint(n, specs)
                 for gene_name, specs in blueprint.items()}
        gene_df = pd.DataFrame(genes)
        pop = Population(blueprint)
        pop.populate.from_genome_dataframe(gene_df)
    else:
        pop = Population(blueprint)
    return pop


def get_random_vals_from_blueprint(n, specs):
    gene_type = specs['type']
    if gene_type == 'numeric':
        return generate_numeric_random_distribution(n, specs)
    elif gene_type == 'categorical':
        return generate_categorical_random_distribution(n, specs)
    elif gene_type == 'vectorial':
        return generate_vectorial_random_distribution(n, specs)
    else:
        raise Exception(
            'Gene type must be "numeric", "categorical" or "vectorial"'
            f' but instead got {gene_type}.')


def generate_numeric_random_distribution(n, specs):
    domain = specs['domain']
    start = specs['ranges']['min']
    stop = specs['ranges']['max']
    progression = specs['ranges']['progression']
    dtype = get_dtype(domain)
    if progression == 'linear':
        return random_linear_dist(start, stop, n, dtype)
    elif progression == 'log':
        return random_log_dist(start, stop, n, dtype)
    else:
        raise Exception(
            'Gene progression must be "linear" or "log"'
            f' but instead got {progression}.')


def generate_categorical_random_distribution(n, specs):
    domain = specs['domain']
    return np.random.choice(domain, n)


def generate_vectorial_random_distribution(n, specs):
    domain = specs['domain']
    replace = specs['ranges']['replace']
    length = specs['ranges']['length']
    return make_vector_list(n, length, domain, replace)



from lamarck.utils import get_id


class Creature:
    def __init__(self, genome=None):
        self.id = None
        self.genome = None
        self.set_genome(genome)

    def __repr__(self):
        return f'Creature <{self.id}> - genome: {self.genome}'

    def set_genome(self, genome):
        self.genome = genome
        self._genomevals = tuple(genome.values())
        self.id = get_id(self.genome)



import numpy as np
from lamarck.utils import random_linear, random_log


class Repopulator:
    def __init__(self, population):
        self.tournament = Tournament(population)
        self.elitism = Elitism(population)
        self.mutate = Mutation(population)


class Reproduce:
    def __init__(self, population):
        self._pop = population

    @property
    def _genome_blueprint(self): return self._pop.genome_blueprint

    def _get_parents(self, idx1, idx2):
        id1 = self._pop.datasets.fitness.index[idx1]
        id2 = self._pop.datasets.fitness.index[idx2]
        parent1 = self._pop.get_creature.from_id(id1)
        parent2 = self._pop.get_creature.from_id(id2)
        return parent1, parent2

    def _populate(self, genomes):
        self._pop.populate.from_genome_list(list(genomes))

    def _cross_over(self, pgenome1, pgenome2):
        flag = True
        while flag:
            schild1 = {}
            schild2 = {}
            for gene, spec in self._genome_blueprint.items():
                pg1 = pgenome1[gene]
                pg2 = pgenome2[gene]
                if spec['type'] == 'vectorial':
                    if spec['ranges']['replace']:
                        g1, g2 = cross_vectorial_genes(pg1, pg2)
                    else:
                        g1, g2 = cross_unique_vectorial_genes(pg1, pg2)
                else:
                    g1, g2 = cross_genes(pg1, pg2)
                schild1.update({gene: g1})
                schild2.update({gene: g2})
            flag = schild1 == schild2
        return schild1, schild2


def cross_genes(pg1, pg2):
    ls = [pg1, pg2]
    i = np.random.randint(0, 2)
    g1 = ls[i]  # pylint: disable=invalid-sequence-index
    g2 = ls[1 - i]
    return g1, g2


def cross_vectorial_genes(pg1, pg2):
    idx_piece = np.random.choice(a=[True, False], size=len(pg1))
    g1 = tuple(np.array(pg1)[idx_piece])
    g2 = tuple(np.array(pg2)[~idx_piece])
    return g1, g2


def cross_unique_vectorial_genes(pg1, pg2):
    cutpoint = np.random.randint(1, len(pg1))
    g1 = pg1[0:cutpoint]
    g2 = pg2[0:cutpoint]
    rem1 = [v for v in pg2 if v not in g1]
    rem2 = [v for v in pg1 if v not in g2]
    remlen = len(pg1) - cutpoint
    g1 += tuple(rem1[0:remlen])
    g2 += tuple(rem2[0:remlen])
    return g1, g2


class Tournament(Reproduce):
    def __call__(self, n_dispute=2, n_children=None, p_children=None):
        """
        Generate offspring by Tournament selection.

        The `Tournament` selection consists of rounding up :n_dispute: Creatures
        and selecting the fittest of the bunch as one of the Parents, and then
        doing the same process for the second parent.

        Parameters
        ----------
        :n_dispute:     Number of Creatures to dispute each of the two `parent`
                        positions. A bigger number means a less random and more
                        elitist parenting selection (default: 2).
        :n_children:    `int` The total amount of children generated. If None is set,
                        the algorithm will keep generating children until the
                        "natural population level" is reached (default: None).
        :p_children:    `float` Substitute for the :n_children: param, if the
                        intention is to provide the number in terms of a proportion
                        of the Population's current size (:n_children: is
                        prioritized).
        """
        n_potential_parents = len(self._pop)
        n_final_pop = get_n_final_pop(self._pop, n_children, p_children)
        n = n_final_pop - len(self._pop)
        genome_list = []
        for _ in range(int(n/2)+1):
            idx1, idx2 = get_parents_min_index(n_potential_parents,
                                               n_dispute)
            parent1, parent2 = self._get_parents(idx1, idx2)
            child_genomes = self._cross_over(parent1.genome,
                                             parent2.genome)
            genome_list += list(child_genomes)
        self._populate(genome_list)


def get_parents_min_index(n_pop, n_dispute):
    idx1 = min(np.random.choice(n_pop, n_dispute, replace=False))
    idx2 = idx1
    while idx2 == idx1:
        idx2 = min(np.random.choice(n_pop, n_dispute, replace=False))
    return idx1, idx2


class Elitism(Reproduce):
    def __call__(self, n_children=None, p_children=None):
        """
        Generate offspring by Elitism selection.

        The `Elitism` selection consists of selecting the top Creatures as the
        most recurring parents, in "best-status-combination" scenarios.

        Parameters
        ----------
        :n_children:    `int` The total amount of children generated. If None is set,
                        the algorithm will keep generating children until the
                        "natural population level" is reached (default: None).
        :p_children:    `float` Substitute for the :n_children: param, if the
                        intention is to provide the number in terms of a proportion
                        of the Population's current size (:n_children: is
                        prioritized).
        ----
        couple 01:  creature01 + creature02     ("sum" = 03, "top" = 01)
        couple 02:  creature01 + creature03     ("sum" = 04, "top" = 01)
        couple 03:  creature01 + creature04     ("sum" = 05, "top" = 01)
        couple 04:  creature02 + creature03     ("sum" = 05, "top" = 02)
        couple 05:  creature01 + creature05     ("sum" = 06, "top" = 01)
        couple 06:  creature02 + creature04     ("sum" = 06, "top" = 02)
        couple 07:  creature01 + creature06     ("sum" = 07, "top" = 01)
        couple 08:  creature02 + creature05     ("sum" = 07, "top" = 02)
        couple 09:  creature03 + creature04     ("sum" = 07, "top" = 03)
        couple 10:  creature01 + creature07     ("sum" = 08, "top" = 01)
        couple 11:  creature02 + creature06     ("sum" = 08, "top" = 02)
        couple 12:  creature03 + creature05     ("sum" = 08, "top" = 03)

        Recurrence (taking this example):
            - creature01:     6 times
            - creature02:     5 times
            - creature03:     4 times
            - creature04:     3 times
            - creature05:     3 times
            - creature06:     2 times
            - creature07:     1 times
        """
        n_final_pop = get_n_final_pop(self._pop, n_children, p_children)
        n = n_final_pop - len(self._pop)
        parent_generator = generate_elite_parent_pair()
        genome_list = []
        for _ in range(int(n/2)+1):
            idx1, idx2 = next(parent_generator)
            parent1, parent2 = self._get_parents(idx1, idx2)
            child_genomes = self._cross_over(parent1.genome,
                                             parent2.genome)
            genome_list += list(child_genomes)
        parent_generator.close()
        self._populate(genome_list)


# I know this code (the next two functions) is filthy but this turned out to be
# a fairly hard problem and I had to keep going...
# I might embellish this later but fwiw this horrible thing is actually working =)
# (Also thank you StackOverflow for the second function =D)

def generate_elite_parent_pair():
    s = 1
    ls = [0, 1]
    subsum = subset_sum(ls, s)
    while True:
        try:
            crs = next(subsum)
            if len(crs) == 2:
                yield crs
        except GeneratorExit:
            return
        except:
            s += 1
            ls.append(ls[-1] + 1)
            subsum = subset_sum(ls, s)


def subset_sum(numbers, target, partial=[], partial_sum=0):
    if partial_sum == target:
        yield tuple(partial)
    if partial_sum >= target:
        return
    for i, n in enumerate(numbers):
        remaining = numbers[i + 1:]
        yield from subset_sum(remaining, target, partial + [n], partial_sum + n)


class Mutation(Reproduce):
    def __call__(self, p, n_genes=1):
        """
        Generates new creatures by randomly picking Creatures based on the
        :mutation probability (p): and randomly changing :n_genes: gene(s).

        Parameters
        ----------
        :p:         `float` Mutation probability.
        :n_genes:   `int` Number of genes tha will be altered (default: 1).
        """
        n = len(self._pop)
        f_creatures = np.random.rand(n) < p
        if any(f_creatures):
            to_be_mutated = self._pop.datasets.input[f_creatures]
            genome_list = [mutate(x[1].to_dict(),
                                  n_genes,
                                  self._genome_blueprint)
                           for x in to_be_mutated.iterrows()]
            self._populate(genome_list)


def mutate(genome, n, blueprint):
    genes = list(genome.keys())
    idxs = np.random.choice(range(len(genes)), replace=False, size=n)
    for idx in idxs:
        gene = genes[idx]
        val = genome[gene]
        specs = blueprint[gene]
        if specs['type'] == 'numeric':
            new_val = mutate_numeric(val, specs)
        elif specs['type'] == 'categorical':
            new_val = mutate_categorical(val, specs)
        elif specs['type'] == 'vectorial':
            new_val = mutate_vectorial(val, specs)
        genome.update({gene: new_val})
    return genome


def mutate_numeric(val, specs):
    minval = specs['ranges']['min']
    maxval = specs['ranges']['max']
    flag = True
    while flag:
        if specs['ranges']['progression'] == 'linear':
            new_val = random_linear(minval, maxval)
        if specs['ranges']['progression'] == 'log':
            new_val = random_log(minval, maxval)
        if specs['domain'] == 'int':
            new_val = int(new_val)
        flag = new_val == val
    return new_val


def mutate_categorical(val, specs):
    domain = specs['domain']
    flag = True
    while flag:
        new_val = np.random.choice(domain)
        flag = new_val == val
    return new_val


def mutate_vectorial(val, specs):
    domain = specs['domain']
    length = specs['ranges']['length']
    replace = specs['ranges']['replace']
    flag = True
    while flag:
        new_val = np.random.choice(domain, length, replace=replace)
        flag = all(new_val == val)
    return tuple(new_val)


def get_n_final_pop(pop, n_children, p_children=None):
    n_potential_parents = len(pop)
    if n_children is None and p_children is None:
        n_final_pop = pop.natural_population_level
    elif p_children is None:
        n_final_pop = n_potential_parents + n_children
    else:
        n_final_pop = int(n_potential_parents * (1 + p_children))
    return n_final_pop


def construct_straight_genome(genome):
    sgenome = []
    for geneval in genome.values():
        if isinstance(geneval, tuple):
            sgenome += list(geneval)
        else:
            sgenome.append(geneval)
    return np.array(sgenome, dtype=object)


def deconstruct_straight_genome(sgenome, genome_blueprint):
    genome = {}
    i = 0
    for gene, spec in genome_blueprint.items():
        if spec['type'] == 'vectorial':
            n = spec['ranges']['length']
            geneval = tuple(sgenome[i:i+n])
        else:
            n = 1
            geneval = sgenome[i]
        i += n
        genome.update({gene: geneval})
    return genome



from copy import deepcopy
from tqdm import tqdm
import numpy as np

from lamarck.assistant import GenomeCreator
from lamarck.assistant import PopulationCreator
from lamarck import Environment


class Optimizer:
    """
    Single or Multi-Objective Genetic Algorithm.

    >>> opt = Optimizer()

    Basic Steps:
        1. Define the Variable Space by setting up the Genome Blueprint
            >>> opt.genome_creator.add_gene_specs.numeric(**kw) # for numeric variables
            >>> opt.genome_creator.add_gene_specs.categorical(**kw) # for categorical variables
            >>> opt.genome_creator.add_gene_specs.vectorial(**kw) # for vectorial variables

        2. Create a Population
            >>> opt.create_population(n_det, n_rand)

        3. Set the Environment: Define the Process Function
            >>> opt.set_process(my_function)

        4. Run
            - Single Objective:
            >>> simulated_pop = opt.run.single_objective(output, objective, **kw)
            - Multi Objective: Ranked
            >>> simulated_pop = opt.run.multi_objective.ranked(priorities, objectives, **kw)
            - Multi Objective: Pareto
            >>> simulated_pop = opt.run.multi_objective.pareto(outputs, objectives, **kw)
    """

    def __init__(self):
        self.genome_creator = GenomeCreator()
        self.pop = None
        self.simulated_pop = None
        self.env = Environment()
        self.run = Runner(self)
        self.set_genome_blueprint()

    @property
    def genome_blueprint(self):
        if self._gb is not None:
            return self._gb
        else:
            return self.genome_creator.get_genome_blueprint()

    @property
    def add_constraint(self):
        if self.pop is not None:
            return self.pop.apply_fitness.add_constraint

    def create_population(self, n_det=0, n_rand=0):
        """
        Create a Population by mixing evenly spaced and random solution ranges.
        
        For pure evenly spaced values, let n_rand be zero; and for pure random
        values, let n_det be zero.

        Parameters
        ----------
        :n_det:     `int` Number of evenly spaced values (subdivisions). The values
                    are selected according to each Gene's Blueprint.
                    Examples:
                        1. numeric linear: min=10, max=90, n_det=5
                            - values = (10, 30, 50, 70, 90)
                        2. numeric log: min=10, max=1000, n_det=3
                            - values = (10, 100, 1000)
                        3. categorical: domain=('A', 'B', 'C', 'D', 'E'), n_det=3
                            - values = ('A', 'C', 'E')
                        4. vectorial: domain=(1, 2, 3, 4), length=3, replace=True, ndet=4
                            - values = (1, 1, 1),
                                       (2, 1, 1),
                                       (3, 1, 1),
                                       (4, 1, 1)
                        5. vectorial: domain=(1, 2, 3, 4), length=3, replace=False, ndet=4
                            - values = (1, 2, 3),
                                       (2, 1, 3),
                                       (3, 1, 2),
                                       (4, 1, 2)
                    - The Population size will normally be the number n_det to the
                    power of the number of genes, except if some duplicate genomes
                    are generated (e.g. 4 genes and n_det=5 will end up generating
                    5**4=625 different combinations of solutions).

        :n_rand:    `int` Number of randomly generated solutions. The values will
                    be randomly generated and assigned to each Genome individually.
                    - The number 'n_rand' will normally end up being the Population
                    size, except if some duplicate genomes are generated.
        """
        pop_creator = PopulationCreator(self.genome_blueprint)
        self.pop = (pop_creator.create.det(n_det)
                    + pop_creator.create.rand(n_rand))
        self.pop.define()
    
    def set_genome_blueprint(self, genome_blueprint=None):
        """
        Manually set the Genome Blueprint. If `None`, the genome_creator's blueprint
        will be used.
        """
        self._gb = genome_blueprint

    def set_population(self, pop):
        """
        Manually insert a Population to be tested. The Population's Genome Blueprint
        will override the genome_creator's.
        """
        self._gb = deepcopy(pop.genome_blueprint)
        self.pop = pop
        self.pop.define()

    def set_process(self, process):
        """
        Define the Process (Python Function) in which the input parameters will be
        optimized by the algorithm.
        """
        self.env.config.set_process(process)


class Runner:
    def __init__(self, opt):
        self._opt = opt
        self.multi_objective = RunnerMulti(opt)

    def single_objective(self, output, objective,
                         max_generations=20, max_stall=None,
                         p_selection=0.5,
                         p_tournament=0.5, n_dispute=2,
                         p_elitism=0.5,
                         p_mutation=0.05, n_per_gene=1):
        """
        Run an Single-Objective Optimization.

        Parameters
        ----------
        :output:            `str` Name of the Variable to be optimized.
        :objective:         `str` Objective ('max', for "maximize"; 'min', for
                            "minimize").
        :max_generations:   `int` Maximum number of Generations (default: 20).
        :max_stall:         `int` Maximum number of times that the simulation
                            doesn't provide a better solution. If `None` is set, the
                            simulation will necessarily run until the
                            :max_generation: is reached (default: None).
        :p_selection:       `float` proportion of the population that will be
                            selected and will be potential Parents to the next
                            Generation (default: 0.5).
        :p_tournament:      `float` proportion of the population that will be
                            generated by the Tournament Selection Model (default:
                            0.5).
        :n_dispute:         `int` Number of selected Creatures that will compete for
                            one of the two Parent spots at any time in the
                            Tournament Selection (default: 2).
        :p_elitism:         `float` proportion of the population that will be
                            generated by the Elitism Selection Model (default: 0.5).
        :p_mutation:        `float` proportion of the population that will generate
                            new Creatures by mutating :n_per_gene: aspects of its
                            genome (default: 0.05).
        :n_per_gene:        `int` number of genes that are mutated (default: 1)
        """
        pop = self._opt.pop.copy()
        env = self._opt.env
        default_simulation('single', pop, env, output, objective,
                           max_generations, max_stall, p_selection,
                           p_tournament, n_dispute, p_elitism,
                           p_mutation, n_per_gene)
        self._opt.simulated_pop = pop
        return pop


class RunnerMulti:
    def __init__(self, opt):
        self._opt = opt

    def pareto(self, outputs, objectives,
               max_generations=20, max_stall=None,
               p_selection=0.5,
               p_tournament=0.5, n_dispute=2,
               p_elitism=0.5,
               p_mutation=0.05, n_per_gene=1):
        """
        Run a Multi-Objective Optimization using the Pareto Fronts to select the
        best range of solutions by finding their unique set of traits, giving the
        desired Outputs and Objectives.

        Parameters
        ----------
        :outputs:           `list` of Variable Names (`str`) to be optimized.
        :objectives:        `list` of Objectives (`str`); ('max', for "maximize";
                            'min', for "minimize").
        :max_generations:   `int` Maximum number of Generations (default: 20).
        :max_stall:         `int` Maximum number of times that the simulation
                            doesn't provide a better solution. If `None` is set, the
                            simulation will necessarily run until the
                            :max_generation: is reached (default: None).
        :p_selection:       `float` proportion of the population that will be
                            selected and will be potential Parents to the next
                            Generation (default: 0.5).
        :p_tournament:      `float` proportion of the population that will be
                            generated by the Tournament Selection Model (default:
                            0.5).
        :n_dispute:         `int` Number of selected Creatures that will compete for
                            one of the two Parent spots at any time in the
                            Tournament Selection (default: 2).
        :p_elitism:         `float` proportion of the population that will be
                            generated by the Elitism Selection Model (default: 0.5).
        :p_mutation:        `float` proportion of the population that will generate
                            new Creatures by mutating :n_per_gene: aspects of its
                            genome (default: 0.05).
        :n_per_gene:        `int` number of genes that are mutated (default: 1)
        """
        pop = self._opt.pop.copy()
        env = self._opt.env
        default_simulation('pareto', pop, env, outputs, objectives,
                           max_generations, max_stall, p_selection,
                           p_tournament, n_dispute, p_elitism,
                           p_mutation, n_per_gene)
        self._opt.simulated_pop = pop
        return pop

    def ranked(self, priorities, objectives,
               max_generations=20, max_stall=None,
               p_selection=0.5,
               p_tournament=0.5, n_dispute=2,
               p_elitism=0.5,
               p_mutation=0.05, n_per_gene=1):
        """
        Run a Multi-Objective Optimization using the Ranked Priority Method to
        select the best range of solutions by ordering them according to the
        "Priorities" list and its respective Objectives.

        Parameters
        ----------
        :priorities:        `list` of Variable Names (`str`) to be optimized. The
                            order of prioritization follows the order of the
                            Variable list.
        :objectives:        `list` of Objectives (`str`); ('max', for "maximize";
                            'min', for "minimize").
        :max_generations:   `int` Maximum number of Generations (default: 20).
        :max_stall:         `int` Maximum number of times that the simulation
                            doesn't provide a better solution. If `None` is set, the
                            simulation will necessarily run until the
                            :max_generation: is reached (default: None).
        :p_selection:       `float` proportion of the population that will be
                            selected and will be potential Parents to the next
                            Generation (default: 0.5).
        :p_tournament:      `float` proportion of the population that will be
                            generated by the Tournament Selection Model (default:
                            0.5).
        :n_dispute:         `int` Number of selected Creatures that will compete for
                            one of the two Parent spots at any time in the
                            Tournament Selection (default: 2).
        :p_elitism:         `float` proportion of the population that will be
                            generated by the Elitism Selection Model (default: 0.5).
        :p_mutation:        `float` proportion of the population that will generate
                            new Creatures by mutating :n_per_gene: aspects of its
                            genome (default: 0.05).
        :n_per_gene:        `int` number of genes that are mutated (default: 1)
        """
        pop = self._opt.pop.copy()
        env = self._opt.env
        default_simulation('ranked', pop, env, priorities, objectives,
                           max_generations, max_stall, p_selection,
                           p_tournament, n_dispute, p_elitism,
                           p_mutation, n_per_gene)
        self._opt.simulated_pop = pop
        return pop


def default_simulation(kind, pop, env, outputs, objectives, max_generations,
                       max_stall, p_selection, p_tournament, n_dispute,
                       p_elitism, p_mutation, n_per_gene):
    stall = 0
    pbar = tqdm(range(max_generations))
    for _ in pbar:
        env.simulate(pop)
        if kind == 'single':
            pop.apply_fitness.single_objective(outputs, objectives)
        elif kind == 'ranked':
            pop.apply_fitness.multi_objective.ranked(outputs, objectives)
        elif kind == 'pareto':
            pop.apply_fitness.multi_objective.pareto(outputs, objectives,
                                                     p_selection)
        if max_stall is not None:
            stall = compute_stall(pop)
            pbar.set_description(f'Stall: {stall}/{max_stall}   ')
            if stall >= max_stall:
                break
        pop.select(p_selection)
        if p_mutation > 0:
            pop.reproduce.mutate(p_mutation)
        if p_tournament > 0:
            pop.reproduce.tournament(n_dispute, p_children=p_tournament)
        if p_elitism > 0:
            pop.reproduce.elitism(p_children=p_elitism)


def compute_stall(pop):
    best_creatures = pop.get_creature.best()
    best_creatures_ids = np.array([creature.id for creature in best_creatures])
    best_ids_list = []
    for gen in range(pop.generation):
        gen_fitness = pop.datasets.get_generation_from_history(gen)
        gen_ranks = pop.fitness_rank_method(gen_fitness)
        f_best = gen_ranks == 1
        best_gen_ids = gen_ranks[f_best].index
        best_ids_list.append(best_gen_ids)
    comparison = [list_similarity(best_creatures_ids, best_ids) > 0.5
                  for best_ids in best_ids_list]
    return sum(comparison)


def list_similarity(list1, list2):
    n = len(list1)
    return sum([item in list2 for item in list1]) / n



from copy import deepcopy


class GenomeCreator:
    """
    This class is used to assist the constructuion of the Genome Blueprint,
    which is used by the PopulationCreator to, well, create Populations.

    A Genome Blueprint is not very easy to build, so this class helps clarifying
    the confusing configurations necessary to properly instruct the
    PopulationCreator.

    Example of a Genome Blueprint

    genome_blueprint = {
        'x': {
            'type': 'numeric',
            'domain': 'int',
            'ranges': {
                'min': 0,
                'max': 10,
                'progression': 'linear',
            }
        },
        'y': {
            'type': 'categorical',
            'domain': ['A', 'B', 'C'],
        },
        'z': {
            'type': 'vectorial',
            'domain': ['i', 'j', 'k'],
            'ranges': {
                'length': 5,
                'replace': True,
            }
        },
    }
    """

    def __init__(self):
        self.reset_genome_blueprint()
        self.add_gene_specs = GeneCreator(self)

    def reset_genome_blueprint(self):
        self._genome_blueprint = {}

    def get_genome_blueprint(self):
        return deepcopy(self._genome_blueprint)


class GeneCreator:
    """
    Gene Creation Assistant
    """

    def __init__(self, genome_creator):
        self._genome_creator = genome_creator

    def numeric(self, name, min, max, progression, domain):
        """
        Generate gene specs for a `numeric` variable.

        Parameters
        ----------
        :name:          `str` that defines the Gene name
        :min:           Gene's minimum possible value
        :max:           Gene's maximum possible value
        :progression:   The way this numeric variable behaves
                        Options:
                            - 'linear'
                            - 'log'
        :domain:        The Gene's domain
                        Options:
                            - 'int'
                            - 'float'
        """
        specs = {
            'type': 'numeric',
            'domain': domain,
            'ranges': {
                'min': min,
                'max': max,
                'progression': progression,
            }
        }
        self._genome_creator._genome_blueprint.update({name: specs})

    def categorical(self, name, domain):
        """
        Generate gene specs for a `categorical` variable.

        Parameters
        ----------
        :name:          `str` that defines the Gene name
        :domain:        `list` containing the entire Gene's domain (All possible values)
                        Examples:
                            1. domain=['Alpha', 'Beta', 'Gamma']
                            2. domain=[1, 2, 4, 8]
        """
        specs = {
            'type': 'categorical',
            'domain': domain,
        }
        self._genome_creator._genome_blueprint.update({name: specs})

    def vectorial(self, name, length, replace, domain):
        """
        Generate gene specs for a `vectorial` variable.

        Parameters
        ----------
        :name:          `str` that defines the Gene name
        :length:        `int` representing the vector's length
        :replace:   `bool` indicating if the vector's values can appear multiple times (True)
                        or just once (False)
        :domain:        `list` containing the entire Gene's domain (All possible values)
                        Example:
                            1. domain=['Alpha', 'Beta', 'Gamma']
                            2. domain=[1, 2, 4, 8]
        """
        specs = {
            'type': 'vectorial',
            'domain': domain,
            'ranges': {
                'length': length,
                'replace': replace,
            }
        }
        self._genome_creator._genome_blueprint.update({name: specs})



import pandas as pd
import numpy as np
import concurrent
from lamarck import Creature, Population
from lamarck.utils import get_id


class Environment:
    def __init__(self):
        self.config = EnvironmentConfig()

    def simulate(self, obj):
        """
        Simulates a `Creature` or a `Population` that has the same genome as the Process's
        Input Parameters.

        Parameters
        ----------
        :obj:   `Creature` or `Population` object

        Examples
        --------
        >>> env.simulate(creature)
        {'out_1': 10, 'out_2': 16.8}

        >>> env.simulate(population)
        Population with <n> Creatures with genes <gene1>, <gene2>, (...).
        """
        if isinstance(obj, Creature):
            return self._simulate_creature(obj)
        elif isinstance(obj, Population):
            self._simulate_population(obj)
        else:
            raise Exception(":obj: must be a Creature or a Population.")

    def _simulate(self, genome):
        return self.config.process(**genome)

    def _simulate_creature(self, creature):
        return self._simulate(creature.genome)

    def _simulate_population(self, pop):
        if self.config.multi:
            self._simulate_multithread(pop)
        else:
            self._simulate_serial(pop)

    def _simulate_serial(self, pop):
        sim_genomes = get_sim_genomes(pop)
        output_dict = {get_id(genome): self._simulate(genome)
                       for genome in sim_genomes}
        build_output(output_dict, pop)

    def _simulate_multithread(self, pop):
        sim_genomes = get_sim_genomes(pop)
        output_dict = {}

        def func(genome):
            output = self._simulate(genome)
            output_dict.update({get_id(genome): output})

        with concurrent.futures.ThreadPoolExecutor() as executor:
            executor.map(func, sim_genomes)
        build_output(output_dict, pop)


def get_sim_genomes(pop):
    out_cols = pop.datasets._outputcols
    if out_cols is None:
        df = pop.datasets.input
    else:
        f = pop.datasets.output[out_cols].isna().all(axis=1)
        index = pop.datasets.index[f]
        df = pop.datasets.input.loc[index]
    return [x[1].to_dict() for x in df.iterrows()]


def build_output(output_dict, pop):
    if any(output_dict):
        df_out = build_output_df_from_output_dict(output_dict)
        outputs = list(df_out.columns)
        pop.set_outputs(outputs)
        pop.add_output(df_out)


def build_output_df_from_output_dict(output_dict):
    index = list(output_dict.keys())
    outputs = get_outputs(output_dict)
    for data in output_dict.values():
        for output in outputs.keys():
            outputs[output].append(data[output])
    return pd.DataFrame(outputs, index=index, dtype=float)


def get_outputs(output_dict):
    output_sample = output_dict[next(iter(output_dict))]
    return {key: [] for key in output_sample}


class EnvironmentConfig:
    def __init__(self):
        self.multi = False
        self.process = None

    def __repr__(self):
        varlist = ['multi', 'process']
        attrlist = [f'{var}:   {val}' for var, val in self.__dict__.items()
                    if var in varlist]
        return '\n'.join(attrlist)

    def set_multi(self, multi):
        """
        Set Multi Thread simulation (True or False).
        """
        self.multi = multi

    def set_process(self, process):
        """
        Sets the Behaviour of this Environment's nature.

        Parameters
        ----------
        :process:   `function` that MUST have only the Genome Blueprint's Genes
                    as parameters and must return the output as a dictionary

        Example
        -------
        genome_blueprint = {
            'x': {
                'type': 'numeric',
                'domain': 'int',
                'ranges': {
                    'min': 0,
                    'max': 10,
                    'progression': 'linear'}},
            'y': {
                'type': 'numeric',
                'domain': 'float',
                'ranges': {
                    'min': 0,
                    'max': 1,
                    'progression': 'linear'}}

        def func(x, y):
            return {
                'out_1': x * y,
                'out_2': x**y - x
            }

        env = Environment()
        env.set_process(func)

        This way, the Environment will be prepare to simulate any Population that
        has this Genome Blueprint.
        """
        self.process = process



import numpy as np
import pandas as pd
from lamarck.plotter import PopulationPlotterPareto
from lamarck.utils import is_objective_ascending, compare_to_value


class FitnessBuilder:
    def __init__(self, population):
        self._pop = population
        self.multi_objective = MultiObjectiveFitness(self)
        self.reset_constraints()
        self.add_constraint = ConstraintAssistant(self)

    def reset_constraints(self):
        self._constraints = {}

    def set_constraints(self, constraints):
        """
        Set Output Constraints to invalidade outputs that fall off the requirements.

        Parameters
        ----------
        :constraints:   `dict` with the output names as `keys` and the constraints
                        as `values`. The constraints must be tuples with the
                        contraint_type-contraint_value pair. 

        Example
        -------
            - constraints = {
                'x': ('gt', 0),    # x must be Greater Than 0
                'y': ('le', 'x')   # y must be Less or Equal to x
            }

        Available Constraints
        ---------------------
            - 'eq'      EQual to
            - 'neq'     Not EQual to
            - 'lt'      Less Than
            - 'le'      Less than or Equal to
            - 'gt'      Greater Than
            - 'ge'      Greater than or Equal to
        """
        self._constraints = constraints

    def single_objective(self, output, objective):
        """
        Define an Output Variable as the objective and an Objective ('maximize'
        or 'minimize'.)

        Parameters
        ----------
        :output:    Variable name (`str`)
        :objective: Objective (`str`).
                    Objective options:
                    - 'minimize' (or 'min')
                    - 'maximize' (or 'max')
        """
        objective = standardize_objective(objective)
        fconstraints = self.get_constraints_filter()
        fitness_df = get_single_criteria(self._pop, output, fconstraints)
        single_rank = single_rank_deco(output, objective)
        self._pop.set_fitness(fitness_df, [objective], single_rank)

    def get_constraints_filter(self):
        if any(self._constraints):
            output_df = self._pop.datasets.output.copy()
            f = pd.Series(np.ones(len(output_df), dtype=bool),
                          index=self._pop.datasets.index)
            for output, (oper, val) in self._constraints.items():
                if isinstance(val, str):
                    val = output_df[val]
                col = output_df[output]
                f = f & compare_to_value(col, oper, val)
        else:
            f = None
        return f


class MultiObjectiveFitness:
    def __init__(self, fitnessbuilder):
        self._fitnessbuilder = fitnessbuilder
        self._pop = fitnessbuilder._pop

    def ranked(self, priorities, objectives):
        """
        Define a list of Output Variables as the objective and a list of
        Objectives in the respective order.

        Parameters
        ----------
        :priorities:    `list` of Output Variables. The order of the output implies
                        the order of priority, meaning that the first output is the
                        one that is targeted to determine the Creature's Fitness,
                        but in case of a Draw, the next one is considered, and so on.
        :objectives:    `list` of Objectives. A `str` of one objective is allowed,
                        meaning all objectives are the same as the one that was
                        passed.
                        Objective options:
                        - 'minimize' (or 'min')
                        - 'maximize' (or 'max')
        """
        objectives = standardize_objectives(objectives)
        fconstraints = self._fitnessbuilder.get_constraints_filter()
        fitness_df = build_criteria_df(self._pop, priorities, objectives,
                                       fconstraints)
        ranked_rank = ranked_rank_deco(priorities, objectives)
        self._pop.set_fitness(fitness_df, objectives, ranked_rank)

    def pareto(self, outputs, objectives, p=0.5):
        """
        Define a list of Output Variables as the objective and a list of
        Objectives in the respective order. All Output Variables are treated
        equally, and there are no "Best" Creatures here, but fronts of
        creatures that offer different types of unique trade-offs.

        Parameters
        ----------
        :outputs:       `list` of Output Variables.
        :objectives:    `list` of Objectives. A `str` of one objective is allowed,
                        meaning all objectives are the same as the one that was
                        passed.
                        Objective options:
                        - 'minimize' (or 'min')
                        - 'maximize' (or 'max')
        :p:             `float` Proportion of the Population that will be classified.
                        (default: 0.5)
        """
        objectives = standardize_objectives(objectives)
        fconstraints = self._fitnessbuilder.get_constraints_filter()
        criteria_df = build_criteria_df(self._pop, outputs, objectives,
                                        fconstraints)
        criteria_df_norm = normalize(criteria_df)
        fronts = get_pareto_fronts(criteria_df_norm, objectives, p)
        crowd = get_pareto_crowds(criteria_df_norm, fronts)
        fitness_df = criteria_df_norm.assign(front=fronts, crowd=crowd)
        self._pop.set_fitness(fitness_df,
                              objectives=['min', 'max'],
                              rank_method=pareto_rank,
                              fitness_cols=['front', 'crowd'])
        self._pop._set_plotter(PopulationPlotterPareto)


def build_criteria_df(pop, criteria, objectives, fconstraints=None):
    n_criteria = len(criteria)
    objectives = assert_objective_list(objectives, n_criteria)
    criteria_dfs = [get_single_criteria(pop, output, fconstraints, suffix=i)
                    for i, output in enumerate(criteria)]
    return pd.concat(criteria_dfs, axis=1)


def assert_objective_list(objectives, n):
    if isinstance(objectives, str):
        objectives = [objectives] * n
    elif not isinstance(objectives, (list, tuple)):
        raise TypeError(":objectives: must be `list`, `tuple` or `str`")
    return objectives


def get_single_criteria(pop, output, fconstraints=None, suffix=''):
    if fconstraints is None:
        criteria = pop.datasets.output[output]
    else:
        criteria = pop.datasets.output[fconstraints][output]
    data = {f'criteria{suffix}': criteria}
    return pd.DataFrame(data, index=criteria.index)


def standardize_objectives(objectives):
    if isinstance(objectives, str):
        return standardize_objective(objectives)
    else:
        return [standardize_objective(objective) for objective in objectives]


def standardize_objective(objective):
    if objective.lower() in ['min', 'minimize']:
        return 'min'
    elif objective.lower() in ['max', 'maximize']:
        return 'max'
    else:
        raise Exception(":objective: must be either 'min' or 'max'.")


def normalize(df):
    return (df - df.mean()) / df.std()


def get_pareto_fronts(df, objectives, p):
    size = len(df)
    find_dominators = find_dominators_deco(df, objectives)
    dominators = df.apply(find_dominators, axis=1)
    fronts = pd.Series(np.zeros(size), index=df.index, dtype=int)
    lenvals = dominators.map(len)
    front = 0
    threshold = size * (1-p)
    while sum(fronts == 0) >= threshold:
        front += 1
        f = lenvals == 0
        fronts[f] = front
        lenvals[f] = None
        dominator_ids = f[f].index
        get_n_dominators = n_dominators_deco(dominator_ids)
        n_dominators = dominators.map(get_n_dominators)
        lenvals[~f] -= n_dominators[~f]
    fronts[fronts == 0] = front + 1
    return fronts


def get_pareto_crowds(df, fronts):
    frontvals = sorted(fronts.unique())
    crowds = pd.Series(np.zeros(len(df)), index=df.index)
    for front in frontvals:
        f = fronts == front
        crowds[f] = get_crowd(df[f])
    return crowds


def get_crowd(df):
    s = pd.Series(np.zeros(len(df)), index=df.index)
    for _, cs in df.iteritems():
        infval = pd.Series([np.inf])
        si = pd\
            .concat([-infval, cs, infval])\
            .sort_values()
        sfvals = si[2:].values - si[:-2].values
        sf = pd.Series(sfvals, index=si.index[1:-1])
        s += sf
    return s

# ranking methods


def single_rank_deco(output, objective):
    def rank_wrapper(df):
        asc = is_objective_ascending(objective)
        return df[output].rank(method='dense', ascending=asc)
    return rank_wrapper


def ranked_rank_deco(priorities, objectives):
    def rank_wrapper(df):
        ranks = []
        for priority, objective in zip(priorities, objectives):
            asc = is_objective_ascending(objective)
            r = df[priority].rank(method='dense', ascending=asc)
            ranks.append(r)
        rank = ranks[-1]
        for r in ranks[::-1]:
            order = np.ceil(np.log10(r.max()))
            factor = 10**order
            rscore = r * factor + rank
            rank = rscore.rank(method='dense')
        return rank
    return rank_wrapper


def pareto_rank(df):
    r1 = df['front'].rank(method='dense')
    r2 = df['crowd'].rank(method='dense', ascending=False)
    order1 = np.ceil(np.log10(r2.max()))
    factor1 = 10**order1
    return (r1 * factor1 + r2).rank(method='dense')


# decorators

def find_dominators_deco(df, objectives):
    def mapper(x):
        f = np.ones(len(df), dtype=bool)
        for col, objective in zip(df.columns, objectives):
            if objective == 'min':
                check = df[col] < x[col]
            elif objective == 'max':
                check = df[col] > x[col]
            f = f & check
        return df.index[f]
    return mapper


def n_dominators_deco(ids):
    def mapper(x):
        return len(np.intersect1d(ids, x))
    return mapper


# Constraint Assistant

class ConstraintAssistant:
    def __init__(self, fitnessbuilder):
        self._fitnessbuilder = fitnessbuilder

    def equal_to(self, output, value):
        constraint = {output: ('eq', value)}
        self._fitnessbuilder._constraints.update(constraint)

    def not_equal_to(self, output, value):
        constraint = {output: ('neq', value)}
        self._fitnessbuilder._constraints.update(constraint)

    def less_than(self, output, value):
        constraint = {output: ('lt', value)}
        self._fitnessbuilder._constraints.update(constraint)

    def less_or_equal(self, output, value):
        constraint = {output: ('le', value)}
        self._fitnessbuilder._constraints.update(constraint)

    def greater_than(self, output, value):
        constraint = {output: ('gt', value)}
        self._fitnessbuilder._constraints.update(constraint)

    def greater_or_equal(self, output, value):
        constraint = {output: ('ge', value)}
        self._fitnessbuilder._constraints.update(constraint)
